/************************************************************************
 * Easings.js
 ************************************************************************
 * Copyright (c) 2021 Pedro Tonini Rosenberg Schneider.
 *
 * This file is part of Pandora.
 *
 * Pandora is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Pandora is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *     
 * You should have received a copy of the GNU General Public License     
 * along with Pandora.  If not, see <https://www.gnu.org/licenses/>.
 *************************************************************************/

/**
 *
 * Derived from Robert Penner's easing equations: http://robertpenner.com/easing/
 * The original copyright notice is as follows:
 * 
 * TERMS OF USE - EASING EQUATIONS
 * 
 * Open source under the BSD License.
 * 
 * Copyright © 2001 Robert Penner
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list of
 * conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, this list
 * of conditions and the following disclaimer in the documentation and/or other materials
 * provided with the distribution.
 * 
 * Neither the name of the author nor the names of contributors may be used to endorse
 * or promote products derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * This {@code Easings} singleton provides an interface for the engine to utilize
 * various easing equations. Used mostly on the Tween GameObject. User can access this
 * API, but it is not recomended; try using tweens instead.
 * 
 * @author Pedro Schneider
 * 
 * @namespace
 */
const Easings = {
    // * See https://easings.net/ for more details on each type of easing.

    /**
     * Namespace to separate linear easing equations.
     * 
     * @namespace
     */
    Linear:
    {
        /**
         * Calculates a linear easing.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        ease: function(t, b, c, d)
        {
            return b + (c * t / d);
        }
    },

    /**
     * Namespace to separate quadratic easing equations.
     * 
     * @namespace
     */
    Quad:
    {
        /**
         * Calculates a quadratic ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            return c * (t /= d) * t + b;
        },

        /**
         * Calculates a quadratic ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            return -c * (t /= d) * (t - 2) + b;
        },

        /**
         * Calculates a quadratic ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            if ((t /= d / 2) < 1) return c / 2 * t * t + b;
            return -c / 2 * ((--t) * (t - 2) - 1) + b;
        }
    },

    /**
     * Namespace to separate cubic easing equations.
     * 
     * @namespace
     */
    Cubic:
    {
        /**
         * Calculates a cubic ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            return c * (t /= d) * t * t + b;
        },

        /**
         * Calculates a cubic ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            return c * ((t = t / d - 1) * t * t + 1) + b;
        },

        /**
         * Calculates a cubic ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            if ((t /= d / 2) < 1) return c / 2 * t * t * t + b;
            return c / 2 * ((t -= 2) * t * t + 2) + b;
        }
    },

    /**
     * Namespace to separate quartic easing equations.
     * 
     * @namespace
     */
    Quart:
    {
        /**
         * Calculates a quartic ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            return c * (t /= d) * t * t * t + b;
        },

        /**
         * Calculates a quartic ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            return -c * ((t = t / d - 1) * t * t * t - 1) + b;
        },

        /**
         * Calculates a quartic ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            if ((t /= d / 2) < 1) return c / 2 * t * t * t * t + b;
            return -c / 2 * ((t -= 2) * t * t * t - 2) + b;
        }
    },

    /**
     * Namespace to separate quintic easing equations.
     * 
     * @namespace
     */
    Quint:
    {
        /**
         * Calculates a quintic ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            return c * (t /= d) * t * t * t * t + b;
        },

        /**
         * Calculates a quintic ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            return c * ((t = t / d - 1) * t * t * t * t + 1) + b;
        },

        /**
         * Calculates a quintic ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            if ((t /= d / 2) < 1) return c / 2 * t * t * t * t * t + b;
            return c / 2 * ((t -= 2) * t * t * t * t + 2) + b;
        }
    },

    /**
     * Namespace to separate senoidal easing equations.
     * 
     * @namespace
     */
    Sine:
    {
        /**
         * Calculates a senoidal ease-in;
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            return -c * Math.cos(t / d * (Math.PI / 2)) + c + b;
        },

        /**
         * Calculates a senoidal ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            return c * Math.sin(t / d * (Math.PI / 2)) + b;
        },

        /**
         * Calculates a senoidal ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            return -c / 2 * (Math.cos(Math.PI * t / d) - 1) + b;
        }
    },

    /**
     * Namespace to separate exponential easing equations.
     * 
     * @namespace
     */
    Expo:
    {
        /**
         * Calculates an exponential ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            return (t == 0) ? b : c * Math.pow(2, 10 * (t / d - 1)) + b;
        },

        /**
         * Calculates an exponential ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            return (t == d) ? b + c : c * (-Math.pow(2, -10 * t / d) + 1) + b;
        },

        /**
         * Calculates an exponential ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            if (t == 0) return b;
            if (t == d) return b + c;
            if ((t /= d / 2) < 1) return c / 2 * Math.pow(2, 10 * (t - 1)) + b;
            return c / 2 * (-Math.pow(2, -10 * --t) + 2) + b;
        }
    },

    /**
     * Namespace to separate circular easing equations.
     * 
     * @namespace
     */
    Circ:
    {
        /**
         * Calculates a circular ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            return -c * (Math.sqrt(1 - (t /= d) * t) - 1) + b;
        },

        /**
         * Calculates a circular ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            return c * Math.sqrt(1 - (t = t / d - 1) * t) + b;
        },

        /**
         * Calculates a circular ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            if ((t /= d / 2) < 1) return -c / 2 * (Math.sqrt(1 - t * t) - 1) + b;
            return c / 2 * (Math.sqrt(1 - (t -= 2) * t) + 1) + b;
        }
    },

    /**
     * Namespace to separate elastic easing equations.
     * 
     * @namespace
     */
    Elastic:
    {
        /**
         * Calculates an elastic ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            var s = 1.70158;
            var p = 0;
            var a = c;
            if (t == 0) return b;
            if ((t /= d) == 1) return b + c;
            if (!p) p = d * .3;
            if (a < Math.abs(c))
            {
                a = c;
                var s = p / 4;
            }
            else var s = p / (2 * Math.PI) * Math.asin(c / a);
            return -(a * Math.pow(2, 10 * (t -= 1)) * Math.sin((t * d - s) * (2 * Math.PI) / p)) + b;
        },

        /**
         * Calculates an elastic ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            var s = 1.70158;
            var p = 0;
            var a = c;
            if (t == 0) return b;
            if ((t /= d) == 1) return b + c;
            if (!p) p = d * .3;
            if (a < Math.abs(c))
            {
                a = c;
                var s = p / 4;
            }
            else var s = p / (2 * Math.PI) * Math.asin(c / a);
            return a * Math.pow(2, -10 * t) * Math.sin((t * d - s) * (2 * Math.PI) / p) + c + b;
        },

        /**
         * Calculates an elastic ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            var s = 1.70158;
            var p = 0;
            var a = c;
            if (t == 0) return b;
            if ((t /= d / 2) == 2) return b + c;
            if (!p) p = d * (.3 * 1.5);
            if (a < Math.abs(c))
            {
                a = c;
                var s = p / 4;
            }
            else var s = p / (2 * Math.PI) * Math.asin(c / a);
            if (t < 1) return -.5 * (a * Math.pow(2, 10 * (t -= 1)) * Math.sin((t * d - s) * (2 * Math.PI) / p)) + b;
            return a * Math.pow(2, -10 * (t -= 1)) * Math.sin((t * d - s) * (2 * Math.PI) / p) * .5 + c + b;
        },
    },

    /**
     * Namespace to separate back easing equations.
     * 
     * @namespace
     */
    Back:
    {
        /**
         * Calculates a back ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            var s = 1.70158;
            return c * (t /= d) * t * ((s + 1) * t - s) + b;
        },

        /**
         * Calculates a back ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            var s = 1.70158;
            return c * ((t = t / d - 1) * t * ((s + 1) * t + s) + 1) + b;
        },

        /**
         * Calculates a back ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            var s = 1.70158;
            if ((t /= d / 2) < 1) return c / 2 * (t * t * (((s *= (1.525)) + 1) * t - s)) + b;
            return c / 2 * ((t -= 2) * t * (((s *= (1.525)) + 1) * t + s) + 2) + b;
        }
    },

    /**
     * Namespace to separate bounce easing equations.
     * 
     * @namespace
     */
    Bounce:
    {
        /**
         * Calculates a bounce ease-in.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeIn: function(t, b, c, d)
        {
            return c - Easings.Bounce.easeOut(d - t, 0, c, d) + b;
        },

        /**
         * Calculates a bounce ease-out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeOut: function(t, b, c, d)
        {
            if ((t /= d) < (1 / 2.75))
                return c * (7.5625 * t * t) + b;
            else if (t < (2 / 2.75))
                return c * (7.5625 * (t -= (1.5 / 2.75)) * t + .75) + b;
            else if (t < (2.5 / 2.75))
                return c * (7.5625 * (t -= (2.25 / 2.75)) * t + .9375) + b;
            else
                return c * (7.5625 * (t -= (2.625 / 2.75)) * t + .984375) + b;
        },

        /**
         * Calculates a bounce ease-in and out.
         * 
         * @param {number} t    current time of the easing process.
         * @param {number} b    beginning value of the esasing process.
         * @param {number} c    change in vlaue during the easing process.
         * @param {number} d    total duration of the easing process.
         *  
         * @returns {number}    current value in the easing process.
         */
        easeInOut: function(t, b, c, d)
        {
            if (t < d / 2) return Easings.Bounce.easeIn(t * 2, 0, c, d) * .5 + b;
            return Easings.Bounce.easeOut(t * 2 - d, 0, c, d) * .5 + c * .5 + b;
        }
    },
}